import sys

import click

from metabolights_utils.commands.submission.submission_create_assay import (
    submission_create_assay,
)
from metabolights_utils.commands.submission.submission_delete_assay import (
    submission_delete_assay,
)
from metabolights_utils.commands.submission.submission_describe import (
    submission_describe,
)
from metabolights_utils.commands.submission.submission_download import (
    submission_download,
)
from metabolights_utils.commands.submission.submission_list import submission_list
from metabolights_utils.commands.submission.submission_login import submission_login
from metabolights_utils.commands.submission.submission_upload import submission_upload
from metabolights_utils.commands.submission.submission_validate import (
    submission_validate,
)


@click.group(name="submission")
def submission_cli():
    """Commands to use MetaboLights study submission REST API."""


submission_cli.add_command(submission_list)
submission_cli.add_command(submission_download)
submission_cli.add_command(submission_describe)
submission_cli.add_command(submission_login)
submission_cli.add_command(submission_upload)
submission_cli.add_command(submission_validate)
submission_cli.add_command(submission_create_assay)
submission_cli.add_command(submission_delete_assay)


if __name__ == "__main__":
    if len(sys.argv) == 1:
        submission_cli(["--help"])
    else:
        submission_cli()



import pathlib
from abc import ABC, abstractmethod
from io import IOBase
from typing import List, Union

from metabolights_utils.isatab.reader import InvestigationFileReaderResult
from metabolights_utils.models.isa.common import IsaTable
from metabolights_utils.models.isa.investigation_file import Investigation
from metabolights_utils.tsv.model import TsvAction, TsvActionReport


class InvestigationFileWriter(ABC):
    """ISA investigation tsv file reader abstract class."""

    @abstractmethod
    def write(
        self,
        investigation: Investigation,
        file_buffer_or_path: Union[IOBase, str, pathlib.Path],
        values_in_quotation_mark: bool = False,
        verify_file_after_update: bool = True,
        skip_parser_info_messages: bool = True,
        investigation_module_name: Union[None, str] = None,
        sync_comments_from_fields: bool = True,
    ) -> InvestigationFileReaderResult:
        """Converts investigation file model to isatab investigation format
        and writes to file buffer.
           If file buffer is not defined, it creates/updates a file on file_path.
           At least one of the file_buffer and file_path parameters should be defined.

        Args:
            investigation (Investigation): investigation file object.
            file_buffer (IOBase, str, pathlib.Path): File buffer to read file content.
            io.StringIO, io.TextIOWrapper with open(), etc.
            values_in_quotation_mark (bool, optional): add values in quotation mark.
            Defaults to False.
            verify_file_after_update (bool, optional): Reads investigation file
            if it is True. Defaults to True.
            skip_parser_info_messages (bool, optional): clear INFO messages
            from parser messages. Defaults to True.
            investigation_module_name: Union[None, str] = investigation file module name
            to find domain classes referenced in json schema.
                e.g. 'metabolights_utils.models.isa.investigation_file'
            sync_comments_from_fields: bool: sync comments
                from fields before writing file.
        Returns:
            InvestigationFileReaderResult: Returns updated investigation model and
            parser messages if verify_file_after_update is True.
                If verify_file_after_update is False, return input investigation
                and empty parser message report
        """


class IsaTableFileWriter(ABC):
    @abstractmethod
    def apply_actions(
        self,
        file_buffer_or_path: Union[IOBase, str, pathlib.Path],
        file_sha256_hash: str,
        actions: List[TsvAction],
    ) -> TsvActionReport:
        """Applies

        Args:
            file_buffer_or_path (Union[IOBase, str, pathlib.Path], optional):
            File path, IOBase or pathlib.Path object.
            file_sha256_hash (str): SH256 of the input file.
            If file SHA256 does not match, method returns error.
            actions (List[TsvAction]): List of allowed actions
        Returns:
            TsvActionReport: results of each action.
            If an action result is not success, result message will be available.
        """

    @abstractmethod
    def save_isa_table(
        self,
        file_path: Union[str, pathlib.Path],
        file_sha256_hash: str,
        isa_table: IsaTable,
        read_encoding: str = "utf-8",
        write_encoding: str = "utf-8",
    ) -> TsvActionReport:
        """Applies

        Args:
            file_path (Union[str, pathlib.Path], optional): File path or pathlib.
            Path object.
            file_sha256_hash (str): SH256 of the input file. ""
            If file SHA256 does not match, method returns error.
            isa_table (IsaTable): isa table model
            read_encoding: encoding of the input file. Defaults to "utf-8".
            write_encoding: encoding of the output file. Defaults to "utf-8".
        Returns:
            TsvActionReport: results of each action. ""
            If an action result is not success, result message will be available.
        """


class IsaTabWriterFactory(ABC):
    @abstractmethod
    def get_investigation_file_writer(self) -> InvestigationFileWriter:
        pass

    @abstractmethod
    def get_assay_file_writer(self) -> IsaTableFileWriter:
        pass

    @abstractmethod
    def get_sample_file_writer(self) -> IsaTableFileWriter:
        pass

    @abstractmethod
    def get_assignment_file_writer(self) -> IsaTableFileWriter:
        pass



import logging
import pathlib
import re
import sys
from io import IOBase
from typing import Callable, Dict, List, Tuple, Union

from pydantic import BaseModel
from pydantic.alias_generators import to_snake

from metabolights_utils.isatab.default.parser.common import read_investigation_file
from metabolights_utils.models.isa import investigation_file
from metabolights_utils.models.isa.common import (
    INVESTIGATION_FILE_INITIAL_ROWS,
    INVESTIGATION_FILE_INITIAL_ROWS_SET,
    INVESTIGATION_FILE_SECTION_NAMES,
    INVESTIGATION_FILE_STUDY_ROWS,
    INVESTIGATION_FILE_STUDY_ROWS_SET,
    Comment,
    IsaAbstractModel,
)
from metabolights_utils.models.isa.investigation_file import (
    BaseSection,
    Investigation,
    Study,
)
from metabolights_utils.models.parser.common import ParserMessage
from metabolights_utils.models.parser.enums import ParserMessageType

logger = logging.getLogger(__name__)

model_module_name = investigation_file.__name__


def parse_investigation_file_content(
    parser: Callable,
    file_path: str,
    messages: List[ParserMessage],
    fix_unicode_exceptions: bool = False,
) -> Tuple[Investigation, List[ParserMessage]]:
    try:
        with pathlib.Path(file_path).open("r", encoding="utf-8") as f:
            model = parser(f, file_path, messages=messages)
            logger.debug("%s is loaded.", file_path)
            return model, messages
    except UnicodeDecodeError as ex:
        logger.warning("Unicode decode error for file %s: %s", file_path, str(ex))
        if fix_unicode_exceptions:
            try:
                with pathlib.Path(file_path).open("r", encoding="latin-1") as f:
                    model = parser(f, file_path, messages=messages)
                    message = ParserMessage(
                        short="File is read with latin-1 encoding",
                        type=ParserMessageType.WARNING,
                    )
                    message.detail = "File is read with latin-1 encoding"
                    messages.append(message)
                    return model, messages
            except Exception as ex:
                logger.exception("File read error for %s: %s", file_path, str(ex))
                message = ParserMessage(
                    short="File read exception", type=ParserMessageType.CRITICAL
                )
                message.detail = f"File parse error: {str(ex)}"
                messages.append(message)
                return None, messages
        else:
            logger.exception("File read error for %s: %s", file_path, str(ex))
            message = ParserMessage(
                short="File read exception", type=ParserMessageType.CRITICAL
            )
            message.detail = f"File parse error: {str(ex)}"
            messages.append(message)
            return None, messages

    except Exception as ex:
        logger.exception("File read error for %s: %s", file_path, str(ex))
        message = ParserMessage(
            short="File read exception", type=ParserMessageType.CRITICAL
        )
        message.detail = f"File parse error: {str(ex)}"
        messages.append(message)
        return None, messages


def parse_investigation_from_fs(
    file_path: str, fix_unicode_exceptions: bool = False
) -> Tuple[Union[Investigation, None], List[ParserMessage]]:
    messages: List[ParserMessage] = []
    file = pathlib.Path(file_path)
    basename = pathlib.Path(file).name
    if not file.exists():
        logger.error("File does not exist %s", basename)
        message = ParserMessage(
            short="File does not exist", type=ParserMessageType.CRITICAL
        )
        message.detail = f"File does not exist: {basename}"
        messages.append(message)
        return None, messages
    file_size = file.stat().st_size
    if file_size == 0:
        logger.error("File is empty: %s", basename)
        message = ParserMessage(short="File is empty", type=ParserMessageType.CRITICAL)
        message.detail = f"File is empty: {basename}"
        messages.append(message)
        return None, messages
    return parse_investigation_file_content(
        get_investigation, file_path, messages, fix_unicode_exceptions
    )


def get_investigation(
    file_buffer: IOBase, file_path: str, messages: List[ParserMessage]
):
    if messages is None:
        messages = []
    index_map = {}
    initial_part = []
    study_parts = []
    initial_part_index_map = {}
    study_parts_index_map = []
    study_parts_index_map = []
    ignore_comments = []
    initial_part_comments = {}
    current_study_part_comments = {}
    study_part_comment_list = []
    is_study_part = False
    current_section = ""
    result = None
    if file_buffer and isinstance(file_buffer, IOBase):
        result: Dict[int, Dict[int, str]] = read_investigation_file(
            file_buffer, messages
        )
    else:
        if (
            file_path
            and isinstance(file_path, pathlib.Path)
            or isinstance(file_path, str)
        ):
            with pathlib.Path(file_path).open("r", encoding="utf-8") as f:
                result: Dict[int, Dict[int, str]] = read_investigation_file(f, messages)

    if result is None:
        logger.warning("File %s is not valid", file_path)
        return Investigation()

    tab: Dict[int, Dict[int, str]] = result
    for line in tab.keys():
        if len(tab[line]) == 0:
            logger.warning("Empty line in file %s, line number: %s", str(line))
            messages.append(
                ParserMessage(
                    short="Empty line",
                    type=ParserMessageType.WARNING,
                    detail=f"Empty line in {str(line)}",
                    line=str(line),
                )
            )
            index_map[line] = ""
            continue
        index_string = tab[line][0]
        index_map[line] = index_string

        if (
            index_string not in INVESTIGATION_FILE_INITIAL_ROWS_SET
            and index_string not in INVESTIGATION_FILE_STUDY_ROWS_SET
        ):
            if index_string.startswith("#"):
                logger.debug(
                    "File comment in %s will be ignored. "
                    "Line number %s, starts with %s  ",
                    file_path,
                    str(line),
                    index_string,
                )
                ignore_comments.append(line)
                messages.append(
                    ParserMessage(
                        type=ParserMessageType.INFO,
                        short="File comment line",
                        detail=f"Comment at line {str(line)}:{index_string}",
                    )
                )
                continue
            if index_string.startswith("Comment"):
                if not is_study_part:
                    is_valid, message_detail = add_line_comment(
                        line, index_string, current_section, initial_part_comments, tab
                    )
                else:
                    is_valid, message_detail = add_line_comment(
                        line,
                        index_string,
                        current_section,
                        current_study_part_comments,
                        tab,
                    )
                if not is_valid:
                    logger.error("Invalid ISA comment line at %s", str(line))
                    message = ParserMessage(
                        short="Invalid ISA comment line", type=ParserMessageType.WARNING
                    )
                    message.detail = f"{message_detail}"
                    message.line = str(line)
                    messages.append(message)
                continue
            else:
                messages.append(
                    ParserMessage(
                        short="Invalid line",
                        type=ParserMessageType.ERROR,
                        detail=f"{index_string}",
                    )
                )
                continue

        if re.match(r"^[A-Z][A-Z0-9 ]+[A-Z0-9]$", index_string):
            current_section = index_string
            if index_string != "STUDY":
                continue

        if not current_section:
            message = ParserMessage(
                short="Invalid start section", type=ParserMessageType.ERROR
            )
            message.detail = (
                f"Line {str(line + 1)}: {index_string} is not valid section start"
            )
            messages.append(message)
            message.line = str(line)

        if not is_study_part:
            if index_string == "STUDY":
                is_study_part = True
            elif index_string not in INVESTIGATION_FILE_INITIAL_ROWS_SET:
                message = ParserMessage(
                    short="Unexpected line in the section", type=ParserMessageType.ERROR
                )
                message.detail = f"Unexpected line in {str(line + 1)}"
                message.line = str(line)
                messages.append(message)
                continue
            else:
                initial_part.append(index_string)
                initial_part_index_map[index_string] = line

        if is_study_part:
            new_study_part = False
            if index_string == "STUDY":
                new_study_part = True
                study_part = []
                study_part_index = {}
                study_part_comments = {}
                study_parts_index_map.append(study_part_index)
                study_part_comment_list.append(study_part_comments)
                current_study_part_comments = study_part_comments
                study_parts.append(study_part)

            if index_string not in INVESTIGATION_FILE_STUDY_ROWS_SET:
                message = ParserMessage(
                    short="Unexpected line in the study section",
                    type=ParserMessageType.ERROR,
                )
                message.detail = f"Unexpected line in {str(line + 1)}"
                message.line = str(line)
                messages.append(message)
                continue
            if study_part or new_study_part:
                study_part.append(index_string)
                study_part_index[index_string] = line

    studies = build_studies(
        file_path,
        messages,
        study_parts,
        study_parts_index_map,
        study_part_comment_list,
        tab,
    )

    investigation, _ = build_investigation(
        file_path,
        initial_part_index_map,
        tab,
        studies,
        initial_part_comments,
        messages=messages,
    )
    investigation.sync_fields_from_comments()
    return investigation


def build_studies(
    source, messages, study_parts, study_parts_index_map, study_part_comment_list, tab
):
    studies = []

    if study_parts:
        for i in range(len(study_parts)):
            study, _ = build_study(
                source,
                study_parts_index_map[i],
                tab,
                messages=messages,
                comments_map=study_part_comment_list[i],
            )
            studies.append(study)
    return studies


def add_line_comment(line_index, line, current_section, comments_list, table):
    val = ""
    try:
        result = re.search(r"^Comment ?\[(.+)\]$", line or "")

        if result and result.groups():
            val = result.group(1)
        else:
            return None, "Not a valid comment line"

    except Exception as ex:
        return None, f"Invalid comment line format {line} {str(ex)}"

    if current_section not in comments_list:
        comments_list[current_section] = []
    values = [table[line_index][i] for i in range(1, len(table[line_index]))]

    comments_list[current_section].append((val, values))
    return result, f"Comment Name:'{val}' Value:'{table[line_index][1]}'"


def build_study(
    file_path,
    index_map: dict,
    content: dict,
    comments_map: dict,
    messages: Union[None, List[ParserMessage]] = None,
):
    study: Study = Study()

    missing_rows = [
        x
        for x in INVESTIGATION_FILE_STUDY_ROWS
        if x not in index_map and x not in INVESTIGATION_FILE_SECTION_NAMES
    ]
    if missing_rows:
        message = ParserMessage(
            short=f"Some rows are missing for study: {', '.join(missing_rows)}",
            type=ParserMessageType.CRITICAL,
        )
        message.detail = "Investigation file is not complete."
        messages.append(message)
        return study, messages
    extra_rows = [
        x
        for x in index_map
        if x not in INVESTIGATION_FILE_STUDY_ROWS
        and not x.lower().startswith(
            "comment[" and x not in INVESTIGATION_FILE_SECTION_NAMES
        )
    ]
    if extra_rows:
        row_messages = ", ".join([f"Line {index_map[x]}:{x}" for x in extra_rows])
        message = ParserMessage(
            short=f"Some rows are missing for study: {row_messages}",
            type=ParserMessageType.CRITICAL,
        )
        message.detail = "Investigation file has unexpected rows."
        messages.append(message)

    assign_by_field_name(
        file_path,
        study,
        index_map,
        content,
        comments_map=comments_map,
        messages=messages,
    )
    return study, messages


def build_investigation(
    file_path,
    index_map: dict,
    content: dict,
    studies: List[Study],
    comments_map: dict,
    messages: Union[None, List[ParserMessage]] = None,
):
    messages = messages if messages is not None else []
    investigation: Investigation = Investigation()

    missing_rows = [
        x
        for x in INVESTIGATION_FILE_INITIAL_ROWS
        if x not in index_map and x not in INVESTIGATION_FILE_SECTION_NAMES
    ]
    if missing_rows:
        message = ParserMessage(
            short=f"Some rows are missing for investigation: {', '.join(missing_rows)}",
            type=ParserMessageType.CRITICAL,
        )
        message.detail = "Investigation file is not complete."
        messages.append(message)
        return investigation, messages

    extra_rows = [
        x
        for x in index_map
        if x not in INVESTIGATION_FILE_INITIAL_ROWS
        and not x.lower().startswith(
            "comment[" and x not in INVESTIGATION_FILE_SECTION_NAMES
        )
    ]
    if extra_rows:
        row_messages = ", ".join([f"Line {index_map[x]}:{x}" for x in extra_rows])
        message = ParserMessage(
            short=f"Some rows are missing for study: {row_messages}",
            type=ParserMessageType.CRITICAL,
        )
        message.detail = "Investigation file has unexpected lines."
        messages.append(message)

    assign_by_field_name(
        file_path,
        investigation,
        index_map,
        content,
        comments_map=comments_map,
        messages=messages,
    )
    investigation.studies = studies
    return investigation, messages


def set_value(obj, key, value):
    camel_key = to_snake(key)
    setattr(obj, camel_key, value)


def assign_by_field_name(
    file_path,
    data: BaseModel,
    index_map: dict,
    tab: dict,
    messages: List[ParserMessage],
    prefix="",
    value_index=1,
    comments_map=None,
) -> None:
    if messages is None:
        messages = []
    data_schema = data.model_json_schema()
    if "properties" not in data_schema or not data_schema["properties"]:
        logger.debug("Properties are not defined in %s", file_path)
        message = ParserMessage(
            short="Property is not defined for object", type=ParserMessageType.WARNING
        )
        message.detail = f"{data.__repr_name__} values will not assigned."
        messages.append(message)
        return

    initial_name_prefix = prefix or ""
    section_prefix = ""
    if isinstance(data, BaseSection):
        section: BaseSection = data
        prefix_value = section.isatab_config.section_prefix
        section_prefix = prefix_value.strip() if prefix_value else ""
        initial_name_prefix = (
            section_prefix
            if not initial_name_prefix
            else f"{initial_name_prefix.strip()} {section_prefix}".strip()
        )
    if isinstance(data, BaseSection) and comments_map:
        section: BaseSection = data
        section_header = section.isatab_config.section_header or ""
        if section_header in comments_map and comments_map[section_header]:
            section.comments = []
            for comment in comments_map[section_header]:
                section.comments.append(Comment(name=comment[0], value=comment[1]))

    for key in data_schema["properties"]:
        if (
            key not in data_schema["properties"]
            or not data_schema["properties"][key]
            or (
                "auto_fill" in data_schema["properties"]
                and data_schema["properties"]["auto_fill"]
            )
        ):
            continue

        field_definition = data_schema["properties"][key]

        field_name = ""
        data_type = "invalid"
        ref_object_definition = ""
        name_prefix = initial_name_prefix

        if (
            "inherit_prefix" in field_definition
            and not field_definition["inherit_prefix"]
        ):
            name_prefix = prefix.strip() if prefix else ""

        if "$ref" in field_definition and field_definition["$ref"]:
            data_type = "ref"
            ref_object_definition = field_definition["$ref"]
        elif "allOf" in field_definition and field_definition["allOf"]:
            if (
                "$ref" in field_definition["allOf"][0]
                and field_definition["allOf"][0]["$ref"]
            ):
                data_type = "ref"
                ref_object_definition = field_definition["allOf"][0]["$ref"]
        elif "anyOf" in field_definition and field_definition["anyOf"]:
            primitives = {"string", "int", "float"}
            for field_item in field_definition["anyOf"]:
                if "type" in field_item and field_item["type"] in primitives:
                    data_type = "string"
                    break
        elif "type" in field_definition:
            data_type = field_definition["type"]
        text_multiple_value = False
        if (
            "text_multiple_value" in field_definition
            and field_definition["text_multiple_value"]
        ):
            text_multiple_value = True
        if field_definition and "header_name" in field_definition:
            if (
                "section_prefix" in field_definition
                and field_definition["section_prefix"]
            ):
                field_name = " ".join(
                    [
                        f"{name_prefix}",
                        f"{field_definition['section_prefix']}",
                        f"{field_definition['header_name']}",
                    ]
                ).strip()
            else:
                field_name = f"{name_prefix} {field_definition['header_name']}".strip()
            if data_type == "array" and not text_multiple_value:
                if "items" in field_definition and field_definition["items"]:
                    items = field_definition["items"]
                    if "$ref" in items and items["$ref"]:
                        search_field = field_name
                        if "search_header" in field_definition:
                            # if (
                            #     "section_prefix" in field_definition
                            #     and field_definition["section_prefix"]
                            # ):
                            #     search_field = " ".join(
                            #         [
                            #             name_prefix,
                            #             field_definition["section_prefix"],
                            #             field_definition["search_header"],
                            #         ]
                            #     )
                            # else:
                            search_field = (
                                f"{name_prefix} {field_definition['search_header']}"
                            )

                        item_list = []
                        set_value(data, key, item_list)

                        if search_field not in index_map:
                            logger.debug(
                                "%s is not in file. Skipping %s in %s",
                                search_field,
                                field_name,
                                file_path,
                            )
                            continue
                        index = index_map[search_field]

                        max_index = len(tab[index])
                        ref_class_name = (
                            items["$ref"]
                            .replace("#/definitions/", "")
                            .replace("#/$defs/", "")
                        )
                        ref_class = getattr(
                            sys.modules[model_module_name],
                            ref_class_name,
                        )
                        for i in range(1, max_index):
                            instance = ref_class()
                            item_list.append(instance)
                            assign_by_field_name(
                                file_path,
                                instance,
                                index_map,
                                tab,
                                prefix=field_name,
                                value_index=i,
                                messages=messages,
                                comments_map=comments_map,
                            )

                        count = len(item_list)
                        for i in range(count):
                            instance_item = item_list[count - i - 1]
                            if IsaAbstractModel.is_empty_model(model=instance_item):
                                item_list.pop()
                            else:
                                break
            elif data_type == "array" and text_multiple_value:
                if "items" in field_definition and field_definition["items"]:
                    items = field_definition["items"]
                    if "$ref" in items and items["$ref"]:
                        ref_class_name = (
                            items["$ref"]
                            .replace("#/definitions/", "")
                            .replace("#/$defs/", "")
                        )
                        ref_class = getattr(
                            sys.modules[model_module_name],
                            ref_class_name,
                        )
                        instance = ref_class()
                        set_value(data, key, [instance])
                        assign_by_field_name(
                            file_path,
                            instance,
                            index_map,
                            tab,
                            prefix=field_name,
                            value_index=value_index,
                            messages=messages,
                            comments_map=comments_map,
                        )
                        item_schema = instance.model_json_schema()

                        new_instances = []
                        separator = ";"
                        for item_key in item_schema["properties"]:
                            item_field_definition = item_schema["properties"][item_key]
                            if (
                                not item_field_definition
                                or "auto_fill" not in item_field_definition
                                or not item_field_definition["auto_fill"]
                            ):
                                continue
                            if (
                                "seperator" in item_field_definition
                                and item_field_definition["seperator"]
                            ):
                                separator = item_field_definition["seperator"]
                            attribute_key = to_snake(item_key)
                            item_val = getattr(instance, attribute_key) or ""
                            separated_values = item_val.split(separator)
                            for _ in range(len(separated_values)):
                                instance_count = len(new_instances)
                                values_count = len(separated_values)
                                if instance_count < values_count:
                                    for _ in range(instance_count, values_count):
                                        new_instances.append(ref_class())
                                for j in range(values_count):
                                    set_value(
                                        new_instances[j], item_key, separated_values[j]
                                    )
                        count = len(new_instances)
                        for i in range(len(new_instances)):
                            instance_item = new_instances[count - i - 1]
                            if IsaAbstractModel.is_empty_model(model=instance_item):
                                new_instances.pop()
                            else:
                                break
                        set_value(data, key, new_instances)
                    else:
                        message = ParserMessage(
                            short="Reference item is not valid",
                            type=ParserMessageType.WARNING,
                        )
                        message.detail = f"Reference item is not valid {data_type} "
                        messages.append(message)
                        logger.debug("Reference item is not valid")
                else:
                    message = ParserMessage(
                        short="There is not item", type=ParserMessageType.WARNING
                    )
                    message.detail = "There is not item"
                    messages.append(message)
                    logger.debug("There is not item")
            elif data_type == "string":
                if field_name not in index_map:
                    message = ParserMessage(
                        short="Index is not defined", type=ParserMessageType.WARNING
                    )
                    message.detail = f"{field_name} is not in {file_path}. "
                    messages.append(message)
                    logger.warning("%sis not in %s Skipping", field_name, file_path)
                    set_value(data, key, "")
                    continue
                index = index_map[field_name]
                value = ""
                if len(tab[index]) > value_index:
                    value = tab[index][value_index]
                else:
                    value = ""
                set_value(data, key, str(value).strip('"').strip() or "")
            elif data_type == "ref":
                ref_class_name = ref_object_definition.replace(
                    "#/definitions/", ""
                ).replace("#/$defs/", "")
                ref_class = getattr(sys.modules[model_module_name], ref_class_name)
                instance = ref_class()
                set_value(data, key, instance)
                assign_by_field_name(
                    file_path,
                    instance,
                    index_map,
                    tab,
                    prefix=field_name,
                    value_index=value_index,
                    messages=messages,
                    comments_map=comments_map,
                )
            else:
                if data_type:
                    message = ParserMessage(
                        short="Invalid data type", type=ParserMessageType.ERROR
                    )
                    message.detail = f"Invalid data type {data_type} in {file_path}"
                    messages.append(message)
                else:
                    message = ParserMessage(
                        short=f"Invalid data type in {file_path}",
                        type=ParserMessageType.ERROR,
                    )
                    message.detail = "Data type is not valid"
                    messages.append(message)


if __name__ == "__main__":
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
    )
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    model, messages = parse_investigation_from_fs(
        "tests/test-data/MTBLS60/i_Investigation.txt"
    )



import json
import logging
import os
from typing import List, Set, Tuple, Union

from metabolights_utils.common import sort_by_study_id
from metabolights_utils.models.common import (
    ErrorMessage,
    GenericMessage,
    InfoMessage,
    WarningMessage,
)
from metabolights_utils.models.enums import GenericMessageType
from metabolights_utils.models.metabolights.model import (
    MetabolightsStudyModel,
    StudyFolderMetadata,
)
from metabolights_utils.provider import definitions
from metabolights_utils.provider.ftp.default_ftp_client import (
    DefaultFtpClient,
    FtpFolderContent,
    LocalDirectory,
)
from metabolights_utils.provider.ftp.folder_metadata_collector import (
    FtpFolderMetadataCollector,
)
from metabolights_utils.provider.ftp.model import FtpFiles
from metabolights_utils.provider.local_folder_metadata_collector import (
    LocalFolderMetadataCollector,
)
from metabolights_utils.provider.study_provider import (
    AbstractDbMetadataCollector,
    MetabolightsStudyProvider,
)
from metabolights_utils.provider.utils import (
    is_metadata_file,
    is_metadata_filename_pattern,
)
from metabolights_utils.utils.filename_utils import join_path

logger = logging.getLogger(__name__)


class MetabolightsFtpRepository(DefaultFtpClient):
    def __init__(
        self,
        local_storage_root_path: Union[None, str] = None,
        ftp_server_url: Union[None, str] = None,
        remote_repository_root_directory: Union[None, str] = None,
        local_storage_cache_path: Union[None, str] = None,
    ) -> None:
        self.local_storage_cache_path = local_storage_cache_path
        if not local_storage_cache_path:
            self.local_storage_cache_path = (
                definitions.default_local_repority_cache_path
            )
        self.local_storage_cache_path = join_path(self.local_storage_cache_path)
        logger.debug(
            "local_storage_cache_path is set to path: %s",
            self.local_storage_cache_path,
        )
        self.local_storage_root_path = local_storage_root_path
        if not self.local_storage_root_path:
            self.local_storage_root_path = (
                definitions.default_local_repository_root_path
            )
        self.local_storage_root_path = join_path(self.local_storage_root_path)
        logger.debug(
            "local_storage_root_path is set to path: %s",
            self.local_storage_root_path,
        )
        self.ftp_server_url = ftp_server_url
        if not self.ftp_server_url:
            self.ftp_server_url = definitions.default_ftp_server_url
        logger.debug(
            "ftp_server_url is set to URL: %s",
            self.ftp_server_url,
        )
        self.remote_repository_root_directory = remote_repository_root_directory
        if not self.remote_repository_root_directory:
            self.remote_repository_root_directory = (
                definitions.default_remote_repository_root_directory
            )
        self.remote_repository_root_directory = (
            self.remote_repository_root_directory.replace("\\", "/")
        )
        logger.debug(
            "remote_repository_root_directory is set to directory: %s",
            self.remote_repository_root_directory,
        )
        self.ftp_client = DefaultFtpClient(
            local_storage_root_path=self.local_storage_root_path,
            ftp_server_url=self.ftp_server_url,
            remote_repository_root_directory=self.remote_repository_root_directory,
        )

    def load_study_model(
        self,
        study_id: str,
        local_path: Union[str, None] = None,
        use_only_local_path: bool = False,
        override_local_files: bool = False,
        load_folder_metadata: bool = True,
        rebuild_folder_index_file: bool = False,
        folder_index_file_path: Union[str, None] = None,
        use_study_model_cache: Union[bool, None] = True,
        study_model_file_path: Union[str, None] = None,
        db_metadata_collector: Union[None, AbstractDbMetadataCollector] = None,
    ) -> Tuple[Union[None, MetabolightsStudyModel], List[GenericMessage]]:
        if not study_id or not study_id.strip():
            return None, [ErrorMessage(short="Invalid study_id")]
        study_id = study_id.upper().strip("/")
        if not local_path:
            local_path = self.local_storage_root_path

        target_path = join_path(local_path, study_id)
        if not folder_index_file_path:
            folder_index_file_path = join_path(
                self.local_storage_cache_path,
                study_id,
                "mtbls_index.json",
            )
        model_cache_path = study_model_file_path
        if not study_model_file_path:
            model_cache_path = join_path(
                self.local_storage_cache_path,
                study_id,
                "study_model.json",
            )

        model: Union[None, MetabolightsStudyModel] = None
        messages = []
        if use_study_model_cache:
            try:
                if os.path.exists(model_cache_path):
                    with open(model_cache_path, encoding="utf-8") as f:
                        data = json.load(f)
                        model = MetabolightsStudyModel.model_validate(data)
                        if model.investigation_file_path == "i_Investigation.txt":
                            messages.append(
                                InfoMessage(short="Loaded from cache file.")
                            )
                            return model, messages
                        else:
                            messages.append(
                                WarningMessage(
                                    short="Cache file is not valid. Skipping..."
                                )
                            )
                else:
                    messages.append(
                        InfoMessage(short="There is no study model cache file.")
                    )
            except Exception as ex:
                messages.append(
                    WarningMessage(
                        short=f"Study model cache file load error: {str(ex)}."
                    )
                )
        if use_only_local_path:
            provider = MetabolightsStudyProvider(
                db_metadata_collector=db_metadata_collector,
                folder_metadata_collector=LocalFolderMetadataCollector(),
            )
            model: MetabolightsStudyModel = provider.load_study(
                study_id,
                study_path=target_path,
                connection=None,
                load_assay_files=True,
                load_sample_file=True,
                load_maf_files=True,
                load_folder_metadata=load_folder_metadata,
                calculate_data_folder_size=True,
                calculate_metadata_size=True,
            )
            if model:
                parent = os.path.dirname(model_cache_path)
                os.makedirs(parent, exist_ok=True)
                with open(model_cache_path, "w", encoding="utf-8") as fw:
                    fw.write(model.model_dump_json(indent=4))

            return model, [InfoMessage(short="Loaded from local isa metadata files.")]

        messages: List[GenericMessage] = []
        try:
            result = self.download_study_metadata_files(
                study_id=study_id,
                local_path=local_path,
                metadata_files=None,
                override_local_files=override_local_files,
                delete_unlisted_metadata_files=True,
            )
            if result.success:
                messages.append(
                    InfoMessage(
                        short="Downloaded metadata file with response",
                        detail=f"Response message: {result.code} {result.message}",
                    )
                )

                provider = MetabolightsStudyProvider(
                    db_metadata_collector=db_metadata_collector,
                    folder_metadata_collector=FtpFolderMetadataCollector(
                        client=self.ftp_client,
                        study_id=study_id,
                        folder_index_file_path=folder_index_file_path,
                        rebuild_folder_index_file=rebuild_folder_index_file,
                    ),
                )
                model: MetabolightsStudyModel = provider.load_study(
                    study_id,
                    study_path=target_path,
                    connection=None,
                    load_assay_files=True,
                    load_sample_file=True,
                    load_maf_files=True,
                    load_folder_metadata=load_folder_metadata,
                    calculate_data_folder_size=False,
                    calculate_metadata_size=False,
                )
            else:
                messages.append(
                    ErrorMessage(
                        short="Download metadata file failure.",
                        detail=f"Error message: {result.code} {result.message}",
                    )
                )

        except Exception as ex:
            messages.append(
                ErrorMessage(
                    short="Download metadata file failure. Try -l option if there is a local copy.",
                    detail=str(ex),
                )
            )
        if model:
            parent = os.path.dirname(model_cache_path)
            os.makedirs(parent, exist_ok=True)
            with open(model_cache_path, "w", encoding="utf-8") as fw:
                fw.write(model.model_dump_json(indent=4))
        return model, messages

    def download_study_metadata_files(
        self,
        study_id: str,
        local_path: Union[List[str], None] = None,
        metadata_files: Union[List[str], None] = None,
        override_local_files: bool = False,
        delete_unlisted_metadata_files: bool = True,
    ) -> LocalDirectory:
        if not study_id:
            return LocalDirectory(root_path="", code=400, message="invalid study_id")
        if not local_path:
            local_path = f"{self.local_storage_root_path}/{study_id}"
        local_path = join_path(local_path)
        response = LocalDirectory(root_path=local_path)

        study_id = study_id.upper().strip("/")

        listed_files = []
        requested_files = metadata_files
        if not metadata_files:
            result = self.list_isa_metadata_files(study_id)
            if result and result.success:
                listed_files = set(result.files)
                requested_files = result.files
            else:
                return LocalDirectory(
                    success=False,
                    code=result.code,
                    message=result.message,
                    local_files=result.files,
                    root_path=local_path,
                )
        if requested_files != metadata_files:
            filtered_files = [
                x for x in requested_files if is_metadata_filename_pattern(x)
            ]
            requested_files = filtered_files
        if requested_files and not os.path.exists(local_path):
            os.makedirs(local_path, exist_ok=True)

        response.actions = {f"{study_id}/{x}": "SKIPPED" for x in listed_files}
        messages: List[GenericMessage] = []
        current_file = None
        try:
            for filename in requested_files:
                # new_local_path = os.path.join(local_path, study_id)
                new_relative_file_path = f"{study_id}/{filename}".replace(
                    "\\", "/"
                ).rstrip("/")
                current_file = filename
                self.ftp_client.download_file(
                    relative_file_path=new_relative_file_path,
                    local_path=local_path,
                    local_files=response,
                    override_local_files=override_local_files,
                    skip_files=None,
                    delete_unlisted_local_files=False,
                    keep_local_files=True,
                )
            if delete_unlisted_metadata_files:
                for filename in os.listdir(local_path):
                    if filename not in listed_files:
                        file_path = join_path(local_path, filename)
                        if is_metadata_file(file_path):
                            response.actions[filename] = "DELETED"
                            os.remove(file_path)
            response.success = True
            response.code = 200
            response.message = "Ok"

        except Exception as ex:
            messages.append(
                ErrorMessage(
                    short=f"Download data file {current_file if current_file else ''} failure",
                    detail=f"Error message: {result.message}, {str(ex)}",
                )
            )

        return response

    def download_study_data_files(
        self,
        study_id: str,
        selected_data_files: Union[List[str], None] = None,
        local_path: Union[List[str], None] = None,
        override_local_files: bool = False,
        skip_files: Union[Set[str], None] = None,
        delete_unlisted_local_files: bool = False,
        keep_local_files: Union[Set[str], None] = None,
    ) -> LocalDirectory:
        if not study_id or not study_id.strip():
            return LocalDirectory(
                code=400,
                message="Invalid study_id",
                root_path=local_path,
            )
        if not local_path:
            local_path = f"{self.local_storage_root_path}/{study_id}"

        response = LocalDirectory(root_path=local_path)
        if not selected_data_files:
            selected_data_files = ["FILES"]
        study_id = study_id.upper().strip("/")
        try:
            for file in selected_data_files:
                new_relative_file_path = f"{study_id}/{file}".replace("\\", "/").rstrip(
                    "/"
                )
                self.ftp_client.download_file(
                    relative_file_path=new_relative_file_path,
                    local_path=local_path,
                    local_files=response,
                    override_local_files=override_local_files,
                    skip_files=skip_files,
                    delete_unlisted_local_files=delete_unlisted_local_files,
                    keep_local_files=keep_local_files,
                )
            response.success = True
            response.code = 200
            response.message = "Ok"
        except Exception as ex:
            response.code = 500
            response.message = str(ex)

        return response

    def list_isa_metadata_files(self, study_id: str) -> FtpFiles:
        response: FtpFolderContent = self.list_study_directory(study_id=study_id)
        isa_files = FtpFiles.model_validate(response.model_dump(), from_attributes=True)
        if response and response.success:
            isa_files.files = []
            for file in response.files:
                if is_metadata_filename_pattern(file):
                    isa_files.files.append(file)
        return isa_files

    def list_studies(self) -> Tuple[List[str], List[GenericMessage]]:
        try:
            result = self.ftp_client.list_directory(None)
            if result.success and result.folders:
                folders = [
                    x.base_name
                    for x in result.descriptors
                    if x.is_directory and x.base_name.startswith("MTBLS")
                ]
                folders.sort(key=sort_by_study_id)
                return folders, []
            else:
                return None, ErrorMessage(
                    short="List study folder failure",
                    detail=f"Response {result.code} {result.message}",
                )
        except Exception as ex:
            return None, ErrorMessage(short="List study folder failure", detail=str(ex))

    def list_study_directory(
        self, study_id: str, subdirectory: Union[str, None] = None
    ) -> FtpFolderContent:
        study_id = study_id.upper().strip("/") if study_id else ""

        directory = f"{study_id}/{subdirectory}" if subdirectory else study_id
        return self.ftp_client.list_directory(directory)

    def rebuild_study_folder_content(
        self, study_id: str, folder_index_file_path: Union[str, None] = None
    ) -> Tuple[Union[None, StudyFolderMetadata], List[GenericMessage]]:
        return self.get_study_folder_content(
            study_id=study_id,
            folder_index_file_path=folder_index_file_path,
            rebuild_folder_index_file=True,
        )

    def get_study_folder_content(
        self,
        study_id: str,
        folder_index_file_path: Union[str, None] = None,
        rebuild_folder_index_file: bool = False,
    ) -> Tuple[Union[None, StudyFolderMetadata], List[GenericMessage]]:
        if not study_id:
            return None, [
                GenericMessage(type=GenericMessageType.ERROR, short="Invalid study_id")
            ]

        study_id = study_id.upper().strip("/")

        collector = FtpFolderMetadataCollector(
            client=self.ftp_client,
            study_id=study_id,
            folder_index_file_path=folder_index_file_path,
            rebuild_folder_index_file=rebuild_folder_index_file,
        )

        metadata, messages = collector.get_folder_metadata(study_path=None)

        return metadata, messages



from typing import Dict, List

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.models.common import MetabolightsBaseModel
from metabolights_utils.models.isa.assay_file import AssayFile
from metabolights_utils.models.isa.assignment_file import AssignmentFile
from metabolights_utils.models.isa.investigation_file import Investigation
from metabolights_utils.models.isa.samples_file import SamplesFile
from metabolights_utils.models.parser.common import ParserMessage


class BaseMetabolightsStudyModel(MetabolightsBaseModel):
    version: Annotated[
        str, Field(description="Version of Metabolights Study Model")
    ] = "v1.1"

    investigation_file_path: Annotated[
        str,
        Field(
            description="Relative path of investigation file. e.g., i_Investigation.txt"
        ),
    ] = ""

    investigation: Annotated[
        Investigation, Field(description="Content of i_Investigation.txt file.")
    ] = Investigation()

    samples: Annotated[
        Dict[str, SamplesFile],
        Field(
            description="Dictionary of sample files and their content. "
            "Only one sample file can be referenced in investigation file."
        ),
    ] = {}

    assays: Annotated[
        Dict[str, AssayFile],
        Field(
            description="Study assay files and their contents. "
            "Multiple assay files can be defined in investigation file."
        ),
    ] = {}

    parser_messages: Annotated[
        Dict[str, List[ParserMessage]],
        Field(
            description="An dictionary to store parser message list for each ISA metadata file contains any type of parser message."
        ),
    ] = {}

    referenced_assignment_files: Annotated[
        List[str],
        Field(
            description="Relative paths of all assignment files referenced in assay files."
        ),
    ] = []
    referenced_raw_files: Annotated[
        List[str],
        Field(description="Relative paths of all raw files referenced in assay files."),
    ] = []
    referenced_derived_files: Annotated[
        List[str],
        Field(
            description="Relative paths of all derived files referenced in assay files."
        ),
    ] = []
    folders_in_hierarchy: Annotated[
        List[str],
        Field(
            description="Relative folder paths that contain files referenced in assay files."
        ),
    ] = []

    metabolite_assignments: Annotated[
        Dict[str, AssignmentFile],
        Field(description="Metabolite assignment files and their contents."),
    ] = {}

    tags: Annotated[
        List[str],
        Field(description="Tag values for the study."),
    ] = []



from typing import List, Union

import click

from metabolights_utils.provider import definitions
from metabolights_utils.provider.submission_repository import (
    MetabolightsSubmissionRepository,
)


@click.command(no_args_is_help=True, name="upload")
@click.option(
    "--local_path",
    "-p",
    default=definitions.default_local_submission_root_path,
    help="Local storage root path. Folder will be created if it does not exist.",
)
@click.option(
    "--rest_api_base_url",
    "-u",
    default=definitions.default_rest_api_url,
    help="MetaboLights study submission API base URL.",
)
@click.option(
    "--override_remote_files",
    "-o",
    is_flag=True,
    default=False,
    help="Downloads files and override current local copies. It is valid if there is no use_only_local option",
)
@click.option(
    "--local_cache_path",
    "-x",
    default=definitions.default_local_submission_cache_path,
    help="Path to store cache files of study submission file indices, study models, etc.",
)
@click.option(
    "--credentials_file_path",
    "-c",
    default=definitions.default_local_submission_credentials_file_path,
    help="Path to store cache files of study submission file indices, study models, etc.",
)
@click.option(
    "--user_api_token",
    "-a",
    help="MetaboLights user API token.",
)
@click.argument("study_id")
@click.argument("metadata_files", required=False)
def submission_upload(
    study_id: str = "",
    metadata_files: Union[List[str], None] = None,
    use_only_local: bool = False,
    local_path: Union[None, str] = None,
    local_cache_path: Union[None, str] = None,
    rest_api_base_url: Union[None, str] = None,
    override_remote_files: bool = False,
    credentials_file_path: str = "",
    user_api_token: Union[str, None] = None,
):
    """
    Uploads local files to private FTP and start sync task to update study folder.

    study_id: MetaboLights study accession number / submission id (MTBLSxxxx or REQXXXX).

    files (optional): files will be downloaded. If not specified, downloads all ISA metadata files.
    """
    study_id = study_id.upper().strip()
    client = MetabolightsSubmissionRepository(
        local_storage_root_path=local_path,
        local_storage_cache_path=local_cache_path,
        rest_api_base_url=rest_api_base_url,
        credentials_file_path=credentials_file_path,
    )
    ftp_details, error = client.get_ftp_upload_details(
        study_id=study_id,
        user_api_token=user_api_token,
        rest_api_base_url=rest_api_base_url,
    )
    if not ftp_details:
        click.echo(f"Upload private ftp details for {study_id}: Failed {str(error)}")
        exit(1)
    success, message = client.upload_metadata_files(
        study_id=study_id,
        override_remote_files=override_remote_files,
        metadata_files=metadata_files,
        ftp_server_url=ftp_details.ftp_host,
        ftp_username=ftp_details.ftp_user,
        ftp_password=ftp_details.ftp_password,
        user_api_token=user_api_token,
    )
    if success:
        click.echo(f"Upload private study {study_id} metadata files: Success")
        click.echo(message)
    else:
        click.echo(f"Upload private study {study_id} metadata files: Failed {message}")
        exit(1)
    success, message = client.upload_data_files(
        study_id=study_id,
        local_path=local_path,
        remote_folder_directory=ftp_details.ftp_folder,
        ftp_server_url=ftp_details.ftp_host,
        ftp_username=ftp_details.ftp_user,
        ftp_password=ftp_details.ftp_password,
    )
    if success:
        click.echo(f"Upload private study {study_id} data files: Success")
        click.echo(message)
        click.echo(f"Sync private FTP metadata files for {study_id}.")
        success, error = client.sync_private_ftp_metadata_files(
            study_id=study_id, pool_period=10, retry=10, user_api_token=user_api_token
        )
        if success:
            click.echo(f"Success: Sync private FTP metadata files for {study_id}.")
            click.echo(f"Sync private FTP data files for {study_id}.")
            success, error = client.sync_private_ftp_data_files(
                study_id=study_id,
                pool_period=10,
                retry=10,
                user_api_token=user_api_token,
            )
            if success:
                click.echo(
                    f"Sync private study date files for {study_id} folder: Success"
                )
            else:
                click.echo(
                    f"Failure: Sync private study metadata files for {study_id}: {error}"
                )

        else:
            click.echo(
                f"Failure: Sync private study metadata files for {study_id}. {error}"
            )

    else:
        click.echo(f"Upload private study {study_id}: {message}")



from metabolights_utils.models.isa import (
    assay_file,
    assignment_file,
    common,
    enums,
    investigation_file,
    samples_file,
)

__all__ = [
    "assay_file",
    "assignment_file",
    "common",
    "enums",
    "investigation_file",
    "samples_file",
]



import json

import click

from metabolights_utils.models.metabolights.model import (
    MetabolightsStudyModel,
)
from metabolights_utils.provider.local_folder_metadata_collector import (
    LocalFolderMetadataCollector,
)
from metabolights_utils.provider.study_provider import (
    MetabolightsStudyProvider,
)


@click.command(no_args_is_help=True, name="create")
@click.option(
    "--output_path",
    "-o",
    help="output file path. e.g. /path/to/output.json",
)
@click.argument("study_path")
def model_create(
    study_path: str,
    output_path: str = "",
):
    """
    Validate submitted study and save validation report on local storage.
    NOTE: Data files should be on FILES subfolder.

    study_path: MetaboLights study folder. Folder should contain ISAtab files.

    output: Output JSON file path. If it is not defined, file is created on working directory.

    """
    folder_metadata_collector = LocalFolderMetadataCollector()
    provider = MetabolightsStudyProvider(
        folder_metadata_collector=folder_metadata_collector,
    )
    model: MetabolightsStudyModel = provider.load_study(
        "temp_study_id",
        study_path=study_path,
        connection=None,
        load_assay_files=True,
        load_sample_file=True,
        load_maf_files=True,
        load_folder_metadata=True,
        calculate_data_folder_size=True,
        calculate_metadata_size=True,
    )
    if not output_path and model.investigation.studies:
        output_path = f"./{model.investigation.studies[0].identifier}.json"

    if model.parser_messages:
        for file, messages in model.parser_messages.items():
            if messages:
                click.echo(f"Messages for {file}")
                for message in messages:
                    click.echo("-", message.type, message.short, message.detail)
    json.dump(model.model_dump(by_alias=True), open(output_path, "w"), indent=2)


if __name__ == "__main__":
    model_create(["tests/test-data/MTBLS1", "-o", "./test-outputs/MTBLS1.json"])



from typing import Union

import click

from metabolights_utils.commands.submission.model import (
    StudyResponse,
    SubmittedStudySummary,
)
from metabolights_utils.commands.utils import convert_html_to_plain_text, split_to_lines
from metabolights_utils.provider import definitions
from metabolights_utils.provider.submission_repository import (
    MetabolightsSubmissionRepository,
)


def print_submitted_study(study: SubmittedStudySummary, log):
    log("-" * 120)
    log(f"{study.accession}: {study.title}")
    public_date = study.release_date
    submission_date = study.created_date
    lines = split_to_lines(convert_html_to_plain_text(study.description))
    for line in lines:
        log(f"\t{line}")
    log(f"\nRelease Date\t: {public_date}\nSubmission Date\t: {submission_date}")


@click.command(name="list")
@click.option(
    "--local_path",
    "-p",
    default=definitions.default_local_submission_root_path,
    help="Local storage root path. Folder will be created if it does not exist.",
)
@click.option(
    "--rest_api_base_url",
    "-f",
    default=f"{definitions.default_rest_api_url}",
    help="MetaboLights Rest API base URL.",
)
@click.option(
    "--local_cache_path",
    "-x",
    default=definitions.default_local_repority_cache_path,
    help="Path to store cache files of submission study folder indices, study models, etc.",
)
@click.option(
    "--use_only_local",
    "-l",
    is_flag=True,
    default=False,
    help="Use only current local directory without REST API.",
)
@click.option(
    "--timeout",
    "-t",
    default=10,
    type=int,
    help="Rest API response timeout.",
)
@click.option(
    "--credentials_file_path",
    "-c",
    default=definitions.default_local_submission_credentials_file_path,
    help="Credentials file path.",
)
@click.argument("study_id", required=False)
@click.argument("subdirectory", required=False)
def submission_list(
    study_id: Union[None, str] = None,
    subdirectory: Union[None, str] = None,
    local_path: Union[None, str] = None,
    local_cache_path: Union[None, str] = None,
    rest_api_base_url: Union[None, str] = None,
    use_only_local: bool = False,
    timeout: int = 10,
    credentials_file_path: str = "",
):
    """
    List submitted studies and study folder content. It works for both local and Rest API.

    study_id (optional): MetaboLights study accession number (MTBLSxxxx). List all studies submitted by user, if not specified.

    subdirectory (optional): Subdirectory of submitted study folder to list its content. List the study root folder if not specified..
    """
    client = MetabolightsSubmissionRepository(
        local_storage_root_path=local_path,
        local_storage_cache_path=local_cache_path,
        rest_api_base_url=rest_api_base_url,
        credentials_file_path=credentials_file_path,
    )

    if not study_id:
        studies_data, error = client.list_studies(timeout=timeout)
        if studies_data:
            studies = studies_data.data.copy()
            studies.sort(key=lambda x: x.updated)
            click.echo("Submitted studies by user:")
            for item in studies:
                print_submitted_study(item, log=click.echo)
            click.echo("-" * 120)
        else:
            click.echo(error)
    else:
        study_id = study_id.upper().strip()

        data, error = client.list_study_directory(
            study_id=study_id, subdirectory=subdirectory, timeout=timeout
        )
        if data:
            studies_response = StudyResponse.model_validate(data)
            studies = studies_response.study.copy()
            studies.sort(key=lambda x: x.file)
            search_path = study_id if not subdirectory else f"{study_id}/{subdirectory}"
            click.echo(f"Submission Study Folder: {search_path}")
            for item in studies:
                click.echo(f"  {item.created_at}\t{item.file}")
        else:
            click.echo(error)


if __name__ == "__main__":
    submission_list()



import pathlib
import uuid

from metabolights_utils.tsv import model as actions
from metabolights_utils.tsv.actions.base import BaseTsvAction


class MoveRowTsvAction(BaseTsvAction):
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: actions.TsvMoveRowAction,
        read_encoding: str = "utf-8",
        write_encoding: str = "utf-8",
    ) -> actions.TsvActionResult:
        result: actions.TsvActionResult = actions.TsvActionResult(action=action)
        if action.action_type != actions.TsvActionType.MOVE_ROW:
            result.message = "Action name is not valid"
            return result

        action: actions.TsvMoveRowAction = action
        new_row_index = action.new_row_index

        if new_row_index is None or new_row_index < 0:
            result.message = "There is no new row index"
            return result

        source_index = action.source_row_index

        if source_index is None or source_index < 0:
            result.message = "There is not source row index"
            return result

        if not action.id:
            uuid_value = str(uuid.uuid4().hex)
            action.id = uuid_value
        new_source_index = source_index
        moved_row = None
        if source_index >= new_row_index:
            with source_file_path.open("r", encoding=read_encoding) as source:
                source.readline()
                row_index = -1
                for line in source:
                    row_index += 1
                    if row_index > source_index:
                        break
                    if row_index == source_index:
                        moved_row = line.strip("\n").split("\t")
                        break

            if not moved_row:
                result.message = "The row will be moved is not found"
                return result
            new_source_index += 1

        moved = False
        source_deleted = False
        try:
            with source_file_path.open("r", encoding=read_encoding) as source:
                header_line = source.readline()

                with target_file_path.open("w", encoding=write_encoding) as target:
                    target.write(header_line)
                    row_index = -1
                    for line in source:
                        row_index += 1
                        if row_index == new_row_index and not moved:
                            self.write_row(target, moved_row)
                            moved = True
                            if source_deleted or row_index != source_index:
                                target.write(line)
                                row_index += 1
                        else:
                            if row_index == new_source_index and not source_deleted:
                                source_deleted = True
                                moved_row = line.strip("\n").split("\t")
                                row_index -= 1
                            else:
                                target.write(line)
                    if new_row_index == row_index + 1 and not moved:
                        self.write_row(target, moved_row)
                        moved = True

            if not moved:
                result.message = "Move is failed"
                return result
            result.success = True
        except Exception as exc:
            result.message = f"{str(exc)}"

        return result



import os
from typing import Union

import click

from metabolights_utils.provider import definitions
from metabolights_utils.provider.submission_repository import (
    MetabolightsSubmissionRepository,
)


@click.command(no_args_is_help=True, name="validate")
@click.option(
    "--rest_api_base_url",
    "-u",
    default=definitions.default_rest_api_url,
    help="MetaboLights study submission API base URL.",
)
@click.option(
    "--local_cache_path",
    "-x",
    default=definitions.default_local_submission_cache_path,
    help="Path to store cache files of study submission file indices, study models, etc.",
)
@click.option(
    "--credentials_file_path",
    "-c",
    default=definitions.default_local_submission_credentials_file_path,
    help="Path to store cache files of study submission file indices, study models, etc.",
)
@click.option(
    "--validation_file_path",
    "-v",
    help="Path to store validation file.",
)
@click.argument("study_id")
def submission_validate(
    study_id: str = "",
    rest_api_base_url: Union[None, str] = None,
    local_cache_path: Union[None, str] = None,
    validation_file_path: bool = False,
    credentials_file_path: str = "",
):
    """
    Validate submitted study and save validation report on local storage.

    study_id: MetaboLights study accession number (MTBLSxxxx).

    """
    if not validation_file_path:
        validation_file_path = os.path.join(
            local_cache_path, study_id, f"{study_id}_validation_report.tsv"
        )
    validation_file_path = os.path.realpath(validation_file_path)
    dir_name = os.path.dirname(validation_file_path)
    os.makedirs(dir_name, exist_ok=True)

    study_id = study_id.upper().strip()
    client = MetabolightsSubmissionRepository(
        local_storage_cache_path=local_cache_path,
        rest_api_base_url=rest_api_base_url,
        credentials_file_path=credentials_file_path,
    )
    success, error = client.validate_study(
        study_id,
        validation_file_path=validation_file_path,
    )

    if success:
        with open(validation_file_path, "r", encoding="utf-8") as f:
            validation_report = f.readlines()
            for line in validation_report:
                click.echo(line.strip())
        click.echo(f"Validation report is saved as {validation_file_path}")
    else:
        click.echo(f"Validate study {study_id} error: {error}.")
        exit(1)


if __name__ == "__main__":
    submission_validate(["MTBLS5397"])



import pathlib
import uuid
from typing import Dict, List

from metabolights_utils.tsv import model as actions
from metabolights_utils.tsv.actions.base import BaseTsvAction


class CopyColumnTsvAction(BaseTsvAction):
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: actions.TsvCopyColumnAction,
        read_encoding: str = "utf-8",
        write_encoding: str = "utf-8",
    ) -> actions.TsvActionResult:
        result: actions.TsvActionResult = actions.TsvActionResult(action=action)
        if action.action_type != actions.TsvActionType.COPY_COLUMN:
            result.message = "Action name is not valid"
            return result

        action: actions.TsvCopyColumnAction = action
        source_column_index = action.source_column_index
        source_column_header = action.source_column_header

        columns: Dict[int, str] = action.target_columns if action.target_columns else {}
        selected_row_indices = (
            set(action.selected_row_indices) if action.selected_row_indices else None
        )
        if not columns:
            result.message = "There is not target column index"
            return result

        column_indices: List[int] = list(columns.keys()).copy()
        column_indices.sort()

        if not action.id:
            uuid_value = str(uuid.uuid4().hex)
            action.id = uuid_value

        try:
            with source_file_path.open("r", encoding=read_encoding) as source:
                header_line = source.readline()
                header_names = header_line.strip("\n").split("\t")
                source_column_found: bool = False
                for column_idx, value in enumerate(header_names):
                    if column_idx == source_column_index:
                        if source_column_header != value:
                            result.message = f"Input header name does not match the actual one for index {column_idx}. Expected: {source_column_header}, found: {value}"
                            return result
                        source_column_found = True
                        break

                if not source_column_found:
                    result.message = (
                        f"Source column index is not found: {source_column_index}."
                    )
                    return result
                invalid_targets = []
                for column_idx in columns:
                    if column_idx >= len(header_names) and column_idx < 0:
                        invalid_targets.append(column_idx)

                if invalid_targets:
                    result.message = f"Target column indices are not valid: {', '.join(invalid_targets)}."
                    return result

                with target_file_path.open("w", encoding=write_encoding) as target:
                    self.write_row(target, header_names)
                    row_index = 0
                    for line in source:
                        row = line.strip("\n").split("\t")
                        for column_idx in columns:
                            if (
                                not selected_row_indices
                                or row_index in selected_row_indices
                            ):
                                row[column_idx] = row[source_column_index]

                        self.write_row(target, row)
                        row_index += 1

            result.success = True
        except Exception as exc:
            result.message = f"{str(exc)}"

        return result
